/*
 * ui.c - hlogin(1) user interface functions
 *
 * Copyleft (C) 2022 ~keith <keith@keithhacks.cyou>
 *
 * This file is part of hlogin.
 *
 * hlogin is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, version 3.
 *
 * hlogin is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with hlogin. If not, see <https://www.gnu.org/licenses/>.
 */
#include "config.h"
#include "ui.h"

#include <panel.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>

#include <locale.h>
#include "gettext.h"

#include <unistd.h>
#include <limits.h>
#include <time.h>

int term_rows, term_cols;

/** ui_init - initialize the UI subsystem
 */
void ui_init()
{
	initscr();
	start_color();
	use_default_colors();
	getmaxyx(stdscr, term_rows, term_cols);
	clear();

	init_pair(1, COLOR_WHITE, COLOR_BLUE );
	init_pair(2, COLOR_BLACK, COLOR_WHITE);
	init_pair(3, COLOR_WHITE, COLOR_WHITE);
}

/** ui_win_new - create a window
 */
struct ui_win * ui_win_new(short flags, int x, int y, int cols, int rows)
{
	struct ui_win *w = calloc(1, sizeof(struct ui_win));
	w->x = x;
	w->y = y;
	w->cols = cols;
	w->rows = rows;
	w->win = newwin(rows, cols, y, x);
	w->panel = new_panel(w->win);
	set_panel_userptr(w->panel, (void *) w);
	w->flags = flags;

	return w;
}

/** ui_win_destroy - destroy a window
 * Items in the window are also destroyed.
 * If UI_CANFREE is set, the object is deallocated.
 */
void ui_win_destroy(struct ui_win *w)
{
	struct ui_item *next;
	for (struct ui_item *item = w->first; item != NULL; item = next) {
		next = item->next;
		ui_item_destroy(item);
	}

	del_panel(w->panel);
	delwin(w->win);

	if (w->flags & UI_CANFREE)
		free(w);
}

/** ui_win_focus - bring a window to the top of the stack
 */
void ui_win_focus(struct ui_win *w)
{
	if (!(w->flags & (UI_HIDDEN | UI_NOFOCUS)))
		top_panel(w->panel);
}

/** ui_win_focusp - check if a window is focused
 */
bool ui_win_focusp(struct ui_win *w)
{
	return panel_below(NULL) == w->panel;
}

/** ui_win_get_focus - get the currently focused window
 */
struct ui_win * ui_win_get_focus()
{
	return (struct ui_win *) panel_userptr(panel_below(NULL));
}

/** ui_win_focus_rotate - rotate windows in the stack
 * If UPWARDS is false, the topmost window moves to the bottom.
 * If UPWARDS is true, the bottom window moves to the top.
 */
void ui_win_focus_rotate(bool upwards)
{
	if (!upwards) // put the focused panel at the bottom
		bottom_panel(panel_below(NULL));
	else // focus the bottom panel
		top_panel(panel_above(NULL));
}

/** ui_win_focus_swap - swap windows in the stack
 * Exchanges the topmost window with the window DEPTH places below it.
 */
void ui_win_focus_swap(int depth)
{
	PANEL *p = NULL;
	for (int i = 0; i < depth-1; i++)
		p = panel_below(p);
	PANEL *q = panel_below(p);

	for (int i = 0; i < depth; i++) {
		top_panel(q);
		q = p;
		p = panel_above(p);
	}

	if (p)
		top_panel(p);
}

/** ui_item_new - create an item
 * The ui_text_new, ui_button_new, etc. functions should be used instead.
 */
struct ui_item * ui_item_new(struct ui_win *w, char type, short flags)
{
	struct ui_item *item = calloc(1, sizeof(struct ui_item));
	item->win = w;
	item->type = type;
	item->flags = flags;

	ui_item_reorder(item, w->last);

	return item;
}

/** ui_item_destroy - destroy an item
 * If UI_CANFREE is set, the object is deallocated.
 */
void ui_item_destroy(struct ui_item *item)
{
	// unlink from list
	if (item->prev)
		item->prev->next = item->next;
	else if (item->win->first == item)
		item->win->first = item->next;
	if (item->next)
		item->next->prev = item->prev;
	else if (item->win->last == item)
		item->win->last = item->prev;

	if (item->win->focus == item)
		item->win->focus = (item->prev != NULL) ? item->prev : item->next;

	if (item->flags & UI_CANFREE)
		free(item);
}

/** ui_item_reorder - reorder an item within its window
 * ITEM is re-linked into its window immediately following AFTER. If
 * AFTER is NULL, ITEM becomes the first item in the window.
 * The order of items in between is not affected.
 */
void ui_item_reorder(struct ui_item *item, struct ui_item *after)
{
	// unlink from list
	if (item->prev)
		item->prev->next = item->next;
	else if (item->win->first == item)
		item->win->first = item->next;
	if (item->next)
		item->next->prev = item->prev;
	else if (item->win->last == item)
		item->win->last = item->prev;

	// link into new position & set item->prev->next
	item->prev = after;
	if (item->prev) {
		item->next = after->next;
		item->prev->next = item;
	} else {
		item->next = item->win->first;
		item->win->first = item;
	}

	// establish item->next->prev
	if (item->next)
		item->next->prev = item;
	else
		item->win->last = item;
}

/** ui_item_navigate: move focus to the previous or next focusable item
 * Focus will be moved to the item immediately before or after ITEM,
 * according to the value of BACKWARDS. Items with UI_NOFOCUS set will
 * be skipped. If either end is reached, navigation will wrap around.
 */
void ui_item_navigate(struct ui_item *item, bool backwards)
{
	do {
		if (!backwards)
			if (item->next)
				item = item->next;
			else
				item = item->win->first;
		else
			if (item->prev)
				item = item->prev;
			else
				item = item->win->last;
	} while (item->flags & UI_NOFOCUS);
	item->win->focus = item;
}

void draw_outline(WINDOW *win, int height, int width, int y, int x, attr_t light, attr_t shadow)
{
	wattron(win, A_ALTCHARSET);

	wattron(win, light);
	mvwhline(win, y, x + 1, ACS_HLINE, width - 2);
	mvwvline(win, y + 1, x, ACS_VLINE, height - 2);

	mvwaddch(win, y, x, ACS_ULCORNER);
	mvwaddch(win, y + height - 1, x, ACS_LLCORNER);
	wattroff(win, light);

	wattron(win, shadow);
	mvwhline(win, y + height - 1, x + 1, ACS_HLINE, width - 2);
	mvwvline(win, y + 1, x + width - 1, ACS_VLINE, height - 2);

	mvwaddch(win, y, x + width - 1, ACS_URCORNER);
	mvwaddch(win, y + height - 1, x + width - 1, ACS_LRCORNER);
	wattroff(win, shadow);

	wattroff(win, A_ALTCHARSET);
}
